from itertools import product as cartprod

class Piece:
    ''' Basic functionality of TicTacToe pieces. '''
    pieces = {
        0: '-',    # 0 is hardcoded to be empty piece
        1: 'X',
        2: 'O',
        # 3: 'Y'   You can now just add new pieces to the dictionary and everything will work properly
    }
    playernumber = len(pieces)-1
    
    def __init__(self, code: int = 0) -> None:
        ''' Initialize piece. '''
        if not isinstance(code, int):
            raise ValueError('Piece cant be initialized with non-int.')
        self.name = Piece.pieces[code]
        self.code = code

    @staticmethod
    def reverse(piece: 'Piece') -> 'Piece':
        ''' Return next piece in Piece.pieces. '''
        if not isinstance(piece, Piece):
            raise ValueError('Cant reverse non-piece object')
        if piece.code == Piece.playernumber:
            return Piece(1)
        else:
            return Piece(piece.code+1)

    @staticmethod
    def previous(piece: 'Piece') -> 'Piece':
        ''' Return previous piece in Piece.pieces.'''
        if not isinstance(piece, Piece):
            raise ValueError('Cant reverse non-piece object')
        if piece.code == 1:
            return Piece(Piece.playernumber)
        else:
            return Piece(piece.code-1)

    def __eq__(self, other: 'Piece') -> bool:
        if not isinstance(other, Piece):
            return NotImplemented
        return self.code == other.code

    def __int__(self) -> int:
        return self.code

    def __str__(self) -> str:
        if hasattr(self.name, '__str__'):
            return self.name
        raise TypeError('This piece cannot be converted to str')

class Board:
    '''
    Only basic functionality of Board. There is no use of Tic Tac Toe game
    logic like win/draw, game ending, turn etc. Game logic is implemented in
    child class TicTacToe.
    '''

    def __init__(self, m=3, n=3) -> None:
        ''' Initialize a board. (m, n) are dimensions of the board. n is number of rows, m is number of columns. '''
        self.size = 0
        self.m = m
        self.n = n
        self.board = [[Piece()]*m for _ in range(n)]

    def columns(self) -> list:
        ''' Return columns for checking win. '''
        board = self.board
        return [list(field) for field in zip(*board)]

    def diagonals(self) -> list:
        ''' Return diagonals for checking win. '''
        if self.m >= self.n:
            diagonals1 = []
            for j in range(self.m-self.n+1):
                diag = []
                for i in range(self.n):
                    diag.append(self[i, i+j])
                diagonals1.append(diag)

            diagonals2 = []
            for j in range(self.m-self.n+1):
                diag = []
                for i in range(self.n):
                    diag.append(self[i, self.n-i-1+j])
                diagonals2.append(diag)
        else:
            diagonals1 = []
            for j in range(self.n-self.m+1):
                diag = []
                for i in range(self.m):
                    diag.append(self[i+j, i])
                diagonals1.append(diag)
    
            diagonals2 = []
            for j in range(self.n-self.m+1):
                diag = []
                for i in range(self.m):
                    diag.append(self[i+j, self.m-i-1])
                diagonals2.append(diag)

        return diagonals1 + diagonals2

    def rows(self):
        ''' Return rows for checking win. '''
        return self.board

    def lines(self) -> list:
        ''' Return all possible lines to be winned. '''
        return self.rows() + self.columns() + self.diagonals()

    def flat(self) -> list:
        ''' Return flat format of board. '''
        board = self.board
        return sum(board, [])
    
    def moves_coords(self) -> list:
        ''' Return all the fields on board'''
        return [(i,j) for (i,j) in cartprod(range(self.n), range(self.m))]

    def __getitem__(self, key):

        if not isinstance(key[0], int) or not isinstance(key[1], int):
            raise TypeError('Board index must be int.')

        if not key in self.moves_coords():
            raise ValueError('Board index out of range.')

        return self.board[key[0]][key[1]]

    def __setitem__(self, key, value):

        if not isinstance(key[0], int) or not isinstance(key[1], int):
            raise TypeError('Board index must be int.')

        if not key in self.moves_coords():
            raise IndexError('Board index out of range.')

        if not isinstance(value, Piece):
            raise ValueError('Cant set Board element to non-piece object.')
        
        self.board[key[0]][key[1]] = value

    def __str__(self) -> str:
        printed = ['\n']
        for row in self.board:
            for field in row:
                printed.append(str(field))
            printed.append('\n')
        return ''.join(printed)